import os
import pytest
import itikz
import tempfile
from IPython.display import SVG, Image


@pytest.mark.skipif(not itikz.JINJA2_ENABLED, reason="jinja2 not installed")
def test_load_and_interpolate_jinja2(monkeypatch, tmpdir):
    # Create the parent_tmpl.md template.
    monkeypatch.chdir(tmpdir)
    parent_src = """# Example\n{% block content -%}{% endblock -%}"""
    parent = tmpdir / "parent_tmpl.md"
    parent.write(parent_src)

    # Use it.
    pkg_name = "itikz"
    child_src = ('{% extends "parent_tmpl.md" %}'
                 "{% block content -%}{{pkg_name}}{% endblock -%}""")
    output = itikz.load_and_interpolate_jinja2(child_src, locals())

    assert output == "# Example\nitikz"


def test_get_cwd(mocker, monkeypatch):
    args = mocker.MagicMock()

    # Case 1: No temp_dir implies the no (`None`) overriding.
    args.temp_dir = False
    assert itikz.get_cwd(args) is None

    # Case 2: The environmental variable demands a temp_dir.
    monkeypatch.setenv('ITIKZ_TEMP_DIR', '1')
    assert itikz.get_cwd(args).endswith("itikz")
    monkeypatch.delenv('ITIKZ_TEMP_DIR')

    # Case 2: Use the system temp_dir.
    args.temp_dir = True
    assert itikz.get_cwd(args).endswith("itikz")


def test_parse_args_does_not_sysexit(capsys):
    itikz.parse_args("-h")
    _, err = capsys.readouterr()
    assert err == ""


def test_build_template_args(mocker):

    # Case 1: No overrides.
    args = itikz.parse_args("")[1]
    res = itikz.build_template_args("code", args)
    assert res == dict(src="code", scale=1, extras="")

    # Case 2: Override scale.
    args = itikz.parse_args("--scale 2")[1]
    res = itikz.build_template_args("code", args)
    assert res == dict(src="code", scale=2.0, extras="")

    # Case 3: Override tex packages.
    args = itikz.parse_args("--scale 2 --tex-packages a,b")[1]
    res = itikz.build_template_args("code", args)
    assert res == dict(src="code", scale=2.0, extras=r"\usepackage{a,b}")

    # Case 4: Override tikz libraries.
    invocation = "--scale 2 --tex-packages a,b --tikz-libraries c,d"
    args = itikz.parse_args(invocation)[1]
    res = itikz.build_template_args("code", args)
    assert res == dict(src="code", scale=2.0,
                       extras="\\usepackage{a,b}\n\\usetikzlibrary{c,d}")


RECTANGLE_TIKZ = r"""
\documentclass[tikz]{standalone}
\begin{document}
\begin{tikzpicture}
\draw[fill=blue] (0, 0) rectangle (1, 1);
\end{tikzpicture}
\end{document}
""".strip()


BAD_TIKZ = "HELLO WORLD"


def test_fetch_or_compile_svg_good_input(tmpdir, monkeypatch):
    expected_md5 = "15d53b05d3a27e1545c9a3688be5e3b4"
    res = itikz.fetch_or_compile_svg(RECTANGLE_TIKZ, 'test_', str(tmpdir))

    for ext in 'tex', 'svg':
        path = tmpdir.join("test_{}.{}".format(expected_md5, ext))
        assert os.path.exists(str(path))

    for ext in 'log', 'aux', 'pdf':
        path = tmpdir.join("test_{}.{}".format(expected_md5, ext))
        assert not os.path.exists(str(path))

    assert isinstance(res, SVG)


def test_fetch_or_compile_svg_bad_input(tmpdir, capsys):
    expected_md5 = "361fadf1c712e812d198c4cab5712a79"
    res = itikz.fetch_or_compile_svg(BAD_TIKZ, 'test_', str(tmpdir))

    for ext in 'tex', 'svg', 'log', 'aux', 'pdf':
        path = tmpdir.join("test_{}.{}".format(expected_md5, ext))
        assert not os.path.exists(str(path))

    assert res is None

    _, err = capsys.readouterr()
    assert 'error' in err.lower()
    assert len(err.splitlines()) == 10


def test_fetch_or_compile_svg_bad_input_full_err(tmpdir, capsys):
    expected_md5 = "361fadf1c712e812d198c4cab5712a79"
    res = itikz.fetch_or_compile_svg(BAD_TIKZ, 'test_', str(tmpdir), True)

    for ext in 'tex', 'svg', 'log', 'aux', 'pdf':
        path = tmpdir.join("test_{}.{}".format(expected_md5, ext))
        assert not os.path.exists(str(path))

    assert res is None

    _, err = capsys.readouterr()
    assert 'error' in err.lower()
    assert len(err.splitlines()) > 10


@pytest.fixture
def itikz_magic(mocker, monkeypatch):
    obj = itikz.TikZMagics()
    shell = mocker.MagicMock()
    shell.user_ns = {}
    monkeypatch.setattr(obj, 'shell', shell)
    return obj


def test_magic_print_help(itikz_magic, capsys):
    assert itikz_magic.itikz("--help") is None
    _, err = capsys.readouterr()
    assert err.startswith("usage: %%itikz")


def test_magic_print_help_on_no_input(itikz_magic, capsys):
    assert itikz_magic.itikz('') is None
    _, err = capsys.readouterr()
    assert err.startswith("usage: %%itikz")


def test_magic_cell_usage(itikz_magic):
    expected_md5 = "15d53b05d3a27e1545c9a3688be5e3b4"
    tmp_dir = os.path.join(tempfile.gettempdir(), 'itikz')

    res = itikz_magic.itikz("--temp-dir --file-prefix test_", RECTANGLE_TIKZ)

    for ext in 'tex', 'svg':
        path = os.path.join(tmp_dir, "test_{}.{}".format(expected_md5, ext))
        assert os.path.exists(str(path))

    for ext in 'log', 'aux', 'pdf':
        path = os.path.join(tmp_dir, "test_{}.{}".format(expected_md5, ext))
        assert not os.path.exists(str(path))

    assert isinstance(res, SVG)


def test_magic_line_usage(itikz_magic):
    expected_md5 = "15d53b05d3a27e1545c9a3688be5e3b4"
    tmp_dir = os.path.join(tempfile.gettempdir(), 'itikz')
    itikz_magic.shell.user_ns['env_src'] = RECTANGLE_TIKZ

    res = itikz_magic.itikz("--temp-dir --file-prefix test_ env_src")

    for ext in 'tex', 'svg':
        path = os.path.join(tmp_dir, "test_{}.{}".format(expected_md5, ext))
        assert os.path.exists(str(path))

    for ext in 'log', 'aux', 'pdf':
        path = os.path.join(tmp_dir, "test_{}.{}".format(expected_md5, ext))
        assert not os.path.exists(str(path))

    assert isinstance(res, SVG)


def test_magic_no_simultanous_standalone_and_pic(itikz_magic, capsys):
    res = itikz_magic.itikz("--implicit-pic --implicit-standalone",
                            RECTANGLE_TIKZ)

    assert res is None
    _, err = capsys.readouterr()
    assert err.startswith("Can't use --implicit")


def test_magic_jinja_without_jinja(itikz_magic, capsys, monkeypatch):
    monkeypatch.setattr(itikz, 'JINJA2_ENABLED', False)
    res = itikz_magic.itikz("--as-jinja", RECTANGLE_TIKZ)

    assert res is None
    _, err = capsys.readouterr()
    assert err.startswith("Please install jinja2")


def test_magic_jinja_print_src(itikz_magic, capsys):
    src = "The answer is: {{n}}"
    itikz_magic.shell.user_ns['n'] = 42
    res = itikz_magic.itikz("--as-jinja --print-jinja", src)

    assert res is None
    out, _ = capsys.readouterr()
    assert out.startswith("The answer is: 42")


def test_implicit_pic(itikz_magic):
    src = r"\node[draw] at (0,0) {Hello World};"
    res = itikz_magic.itikz("--implicit-pic --temp-dir", src)
    assert isinstance(res, SVG)


def test_implicit_standalone(itikz_magic):
    pic = r"\node[draw] at (0,0) {Hello World};"
    src = "\\begin{tikzpicture}\n" + pic + "\n\\end{tikzpicture}\n"
    cmd = "--implicit-standalone --tex-packages=tikz --temp-dir"
    res = itikz_magic.itikz(cmd, src)
    assert isinstance(res, SVG)


def test_rasterize_good_input(itikz_magic):
    pic = r"\node[draw] at (0,0) {Hello World};"
    src = "\\begin{tikzpicture}\n" + pic + "\n\\end{tikzpicture}\n"
    cmd = "--implicit-standalone --tex-packages=tikz --temp-dir --rasterize"
    res = itikz_magic.itikz(cmd, src)
    assert isinstance(res, Image)


def test_rasterize_bad_input(itikz_magic):
    cmd = "--temp-dir --rasterize"
    res = itikz_magic.itikz(cmd, BAD_TIKZ)
    assert res is None


def test_rasterize_no_cairo_svg(itikz_magic, monkeypatch, capsys):
    monkeypatch.setattr(itikz, 'CAIROSVG_ENABLED', False)

    pic = r"\node[draw] at (0,0) {Hello World};"
    src = "\\begin{tikzpicture}\n" + pic + "\n\\end{tikzpicture}\n"
    cmd = "--implicit-standalone --tex-packages=tikz --temp-dir --rasterize"
    res = itikz_magic.itikz(cmd, BAD_TIKZ)

    assert res is None
    _, err = capsys.readouterr()
    assert err.startswith("Please install cairosvg")



import argparse
import os
import shlex
import sys
import tempfile
from hashlib import md5
from shutil import fnmatch
from string import Template
from subprocess import check_output, CalledProcessError
from IPython.display import SVG, Image
from IPython.core.magic import Magics, magics_class, line_cell_magic


__author__ = """John Bjorn Nelson"""
__email__ = 'jbn@abreka.com'
__version__ = '0.1.2'


IMPLICIT_PIC_TMPL = Template(r"""\documentclass[tikz]{standalone}
$extras
\begin{document}
\begin{tikzpicture}[scale=$scale]
$src
\end{tikzpicture}
\end{document}""")


IMPLICIT_STANDALONE = Template(r"""\documentclass{standalone}
$extras
\begin{document}
$src
\end{document}""")


JINJA2_ENABLED = True
try:
    import jinja2
except ImportError:  # pragma: no cover
    JINJA2_ENABLED = False


CAIROSVG_ENABLED = True
try:
    import cairosvg
except ImportError:  # pragma: no cover
    CAIROSVG_ENABLED = False


def parse_args(line):
    parser = argparse.ArgumentParser(prog="%%itikz",
                                     description='Tikz to tex to SVG',
                                     add_help=False)

    parser.add_argument('k', type=str, nargs='?',
                        help='the variable in IPython with the string source')

    parser.add_argument('--temp-dir', dest='temp_dir', action='store_true',
                        default=False,
                        help='emit artifacts to system temp dir')

    parser.add_argument('--file-prefix', dest='file_prefix',
                        default='', help='emit artifacts with a path prefix')

    parser.add_argument('--implicit-pic', dest='implicit_pic',
                        action='store_true', default=False,
                        help='wrap source in implicit tikzpicture document')

    parser.add_argument('--implicit-standalone', dest='implicit_standalone',
                        action='store_true', default=False,
                        help='wrap source in implicit document')

    parser.add_argument('--scale', dest='scale',
                        default='1',
                        help='Set tikzpicture scale in --implicit-pic tmpl')

    parser.add_argument('--tikz-libraries', dest='tikz_libraries',
                        default='',
                        help='Comma separated list of tikz libraries to use')

    parser.add_argument('--tex-packages', dest='tex_packages',
                        default='',
                        help='Comma separated list of tex packages to use')

    parser.add_argument('--as-jinja', dest='as_jinja',
                        action='store_true', default=False,
                        help="Interpret the source as a jinja2 template")

    parser.add_argument('--print-jinja', dest='print_jinja',
                        action='store_true', default=False,
                        help="Print interpolated jinja2 source then bail.")

    parser.add_argument('--rasterize', dest='rasterize',
                        action='store_true', default=False,
                        help="Rasterize the svg with cairosvg")

    parser.add_argument('--full-error', dest='full_err',
                        action='store_true', default=False,
                        help="Emit the full error message")

    # Override help: the default does a sys.exit()
    parser.add_argument('-h', '--help', dest='print_help',
                        action='store_true', default=False,
                        help='show this help message')

    return parser, parser.parse_args(shlex.split(line))


def get_cwd(args):
    if args.temp_dir or os.environ.get('ITIKZ_TEMP_DIR'):
        cwd = os.path.join(tempfile.gettempdir(), 'itikz')
        os.makedirs(cwd, exist_ok=True)
        return cwd  # Override as cwd
    else:
        return None  # No override.


def fetch_or_compile_svg(src, prefix='', working_dir=None, full_err=False):
    src_hash = md5(src.encode()).hexdigest()
    output_path = prefix + src_hash
    if working_dir is not None:
        output_path = os.path.join(working_dir, output_path)
    svg_path = output_path + ".svg"

    if not os.path.exists(svg_path):
        tex_path = output_path + ".tex"
        pdf_path = output_path + ".pdf"

        with open(tex_path, "w") as fp:
            fp.write(src)

        try:
            check_output(["pdflatex", tex_path], cwd=working_dir)
            check_output(["pdf2svg", pdf_path, svg_path], cwd=working_dir)
        except CalledProcessError as e:
            cleanup_artifacts(working_dir, src_hash)
            err_msg = e.output.decode()

            if not full_err:  # tail -n 10
                err_msg = "\n".join(err_msg.splitlines()[-10:])

            print(err_msg, file=sys.stderr)
            return

        cleanup_artifacts(working_dir, src_hash, svg_path, tex_path)

    with open(svg_path, "r") as fp:
        return SVG(fp.read())


def cleanup_artifacts(working_dir, src_hash, *retaining):
    glob = "*{}*".format(src_hash)

    for file_name in fnmatch.filter(os.listdir(working_dir), glob):
        file_path = os.path.join(working_dir or '', file_name)
        if file_path not in retaining:
            os.unlink(file_path)


def load_and_interpolate_jinja2(src, ns):
    # The FileSystemLoader should operate in the current working directory.
    # By assumption, extended jinja templates aren't temporary files --
    # the user wrote them by hand. They are part of code you would want in
    # your repository!
    fs_loader = jinja2.FileSystemLoader(os.getcwd())
    tmpl_env = jinja2.Environment(loader=fs_loader)

    # The final template -- the one that may extend a custom template --
    # may be in the current directory or in a temporary one. So, it's
    # passed as a string.
    tmpl = tmpl_env.from_string(src)

    return tmpl.render(**ns)


@magics_class
class TikZMagics(Magics):

    @line_cell_magic
    def itikz(self, line, cell=None):
        src = cell
        parser, args = parse_args(line)

        if args.print_help:
            parser.print_help(file=sys.stderr)
            return

        ipython_ns = self.shell.user_ns

        if cell is None:

            if args.k is None or args.k not in ipython_ns:
                parser.print_usage(file=sys.stderr)
                return

            src = ipython_ns[args.k]

        if args.implicit_pic and args.implicit_standalone:
            print("Can't use --implicit-standalone and --implicit-pic",
                  file=sys.stderr)
            return None

        # Jinja processing comes BEFORE implicit pic or standalone processing!
        if args.as_jinja:

            if not JINJA2_ENABLED:
                print("Please install jinja2", file=sys.stderr)
                print("$ pip install jinja2", file=sys.stderr)
                return

            src = load_and_interpolate_jinja2(src, ipython_ns)

            if args.print_jinja:
                print(src)
                return

        if args.implicit_pic:
            src = IMPLICIT_PIC_TMPL.substitute(build_template_args(src, args))
        elif args.implicit_standalone:
            tmpl_args = build_template_args(src, args)
            src = IMPLICIT_STANDALONE.substitute(tmpl_args)

        svg = fetch_or_compile_svg(src, args.file_prefix, get_cwd(args),
                                   args.full_err)

        if svg is None:
            return None

        if args.rasterize:
            if not CAIROSVG_ENABLED:
                print("Please install cairosvg", file=sys.stderr)
                print("$ pip install cairosvg", file=sys.stderr)
                return

            png_bytes = cairosvg.svg2png(bytestring=svg.data.encode())

            return Image(data=png_bytes)

        return svg




def build_template_args(src, args):
    extras = []

    if args.tex_packages:
        extras.append(r"\usepackage{" + args.tex_packages + "}")

    if args.tikz_libraries:
        extras.append(r"\usetikzlibrary{" + args.tikz_libraries + "}")

    extras = "\n".join(extras)

    return dict(src=src,
                scale=float(args.scale),
                extras=extras)


def load_ipython_extension(ipython):  # pragma: no cover
    ipython.register_magics(TikZMagics)
